@@ -53,13 +53,13 @@ import Cachix.Client.Retry (retryAll)
 import Cachix.Daemon.NarinfoQuery qualified as NarinfoQuery
 import Cachix.Daemon.Protocol qualified as Protocol
 import Cachix.Daemon.PushManager.PushJob qualified as PushJob
+import Cachix.Daemon.TaskQueue
 import Cachix.Daemon.Types.Log (Logger)
 import Cachix.Daemon.Types.PushEvent (PushEvent (..), PushEventMessage (..), newPushRetryStatus)
 import Cachix.Daemon.Types.PushManager
 import Cachix.Types.BinaryCache qualified as BinaryCache
 import Conduit qualified as C
 import Control.Concurrent.Async qualified as Async
-import Control.Concurrent.STM.TBMQueue
 import Control.Concurrent.STM.TVar
 import Control.Monad.Catch qualified as E
 import Control.Monad.IO.Unlift (MonadUnliftIO)
@@ -68,6 +68,7 @@ import Control.Retry (RetryStatus, rsIterNumber)
 import Data.ByteString qualified as BS
 import Data.HashMap.Strict qualified as HashMap
 import Data.IORef
+import Data.Sequence qualified as Seq
 import Data.Set qualified as Set
 import Data.Text qualified as T
 import Data.Time (UTCTime, diffUTCTime, getCurrentTime, secondsToNominalDiffTime)
@@ -86,18 +87,14 @@ newPushManagerEnv pushOptions batchOptions pmPushParams onPushEvent pmLogger = l
   pmPushJobs <- newTVarIO mempty
   pmPendingJobCount <- newTVarIO 0
   pmStorePathIndex <- newTVarIO mempty
-  pmTaskQueue <- newTBMQueueIO 100_000
+  pmTaskQueue <- atomically newTaskQueue
   pmTaskSemaphore <- QSem.newQSem (numJobs pushOptions)
   pmLastEventTimestamp <- newTVarIO =<< getCurrentTime
   let pmOnPushEvent id pushEvent = updateTimestampTVar pmLastEventTimestamp >> onPushEvent id pushEvent
 
   -- Create query manager with callback that queues ProcessQueryResponse tasks
   let batchCallback requestId response = do
-        atomically $ do
-          result <- tryWriteTBMQueue pmTaskQueue $ HandleMissingPathsResponse requestId response
-          case result of
-            Just True -> return ()
-            _ -> retry -- Queue is full, keep trying
+        atomically $ writeTask pmTaskQueue $ HandleMissingPathsResponse requestId response
   pmNarinfoQueryManager <- NarinfoQuery.new batchOptions batchCallback
 
   return $ PushManagerEnv {..}
@@ -110,7 +107,7 @@ stopPushManager timeoutOptions PushManagerEnv {..} = do
   atomicallyWithTimeout timeoutOptions pmLastEventTimestamp $ do
     pendingJobs <- readTVar pmPendingJobCount
     check (pendingJobs <= 0)
-  atomically $ closeTBMQueue pmTaskQueue
+  atomically $ closeTaskQueue pmTaskQueue
 
 -- | Start the batch processor for narinfo queries
 startBatchProcessor :: (MonadUnliftIO m, Katip.KatipContext m) => PushManagerEnv -> m ()
@@ -141,12 +138,12 @@ addPushJob pushRequest = do
   let queueJob = do
         modifyTVar' pmPushJobs $ HashMap.insert (pushId pushJob) pushJob
         incrementTVar pmPendingJobCount
-        res <- tryWriteTBMQueue pmTaskQueue $ QueryMissingPaths (pushId pushJob)
+        res <- tryWriteTask pmTaskQueue $ QueryMissingPaths (pushId pushJob)
         case res of
           Just True -> return True
-          _ -> retry
+          _ -> return False
 
-  didQueue <- liftIO $ atomically $ queueJob `orElse` return False
+  didQueue <- liftIO $ atomically queueJob
 
   if didQueue
     then return $ Just $ pushId pushJob
@@ -195,7 +192,7 @@ modifyPushJobSTM pushJobs pushId f =
     let pj = HashMap.adjust f pushId jobs
     (HashMap.lookup pushId pj, pj)
 
-modifyPushJobs :: [Protocol.PushRequestId] -> (PushJob -> PushJob) -> PushManager ()
+modifyPushJobs :: (Foldable f) => f Protocol.PushRequestId -> (PushJob -> PushJob) -> PushManager ()
 modifyPushJobs pushIds f = do
   pushJobs <- asks pmPushJobs
   liftIO $ atomically $ modifyTVar' pushJobs $ \pushJobs' ->
@@ -223,9 +220,9 @@ queueStorePaths pushId storePaths = do
   let addToQueue storePath = do
         isDuplicate <- HashMap.member storePath <$> readTVar pmStorePathIndex
         unless isDuplicate $
-          writeTBMQueue pmTaskQueue (PushPath storePath)
+          writeTask pmTaskQueue (PushPath storePath)
 
-        modifyTVar' pmStorePathIndex $ HashMap.insertWith (<>) storePath [pushId]
+        modifyTVar' pmStorePathIndex $ HashMap.insertWith (<>) storePath (Seq.singleton pushId)
 
   transactionally $ map addToQueue storePaths
 
@@ -235,11 +232,11 @@ removeStorePath storePath = do
   liftIO $ atomically $ do
     modifyTVar' storePathIndex $ HashMap.delete storePath
 
-lookupStorePathIndex :: FilePath -> PushManager [Protocol.PushRequestId]
+lookupStorePathIndex :: FilePath -> PushManager (Seq.Seq Protocol.PushRequestId)
 lookupStorePathIndex storePath = do
   storePathIndex <- asks pmStorePathIndex
   references <- liftIO $ readTVarIO storePathIndex
-  return $ fromMaybe [] (HashMap.lookup storePath references)
+  return $ fromMaybe Seq.empty (HashMap.lookup storePath references)
 
 checkPushJobCompleted :: Protocol.PushRequestId -> PushManager ()
 checkPushJobCompleted pushId = do
@@ -483,7 +480,7 @@ pushFinished pushJob@PushJob {pushId} = void $ runMaybeT $ do
 
   lift $ removePushJob pushId
 
-sendStorePathEvent :: [Protocol.PushRequestId] -> PushEventMessage -> PushManager ()
+sendStorePathEvent :: (Foldable f) => f Protocol.PushRequestId -> PushEventMessage -> PushManager ()
 sendStorePathEvent pushIds msg = do
   timestamp <- liftIO getCurrentTime
   sendPushEvent <- asks pmOnPushEvent
